module ClusterEnsembles

    using LinearAlgebra
    using Metis
    using LightGraphs
    using SimpleWeightedGraphs
    using Distances
    using Clustering
    using Random
    using SparseArrays
    using KaHyPar

    const default_configuration = joinpath(@__DIR__, "./config/cut_kKaHyPar_sea20.ini")

    export cluster_ensembles

    include("metis_patch.jl")

    function set_nclass(labels)
        nclass = -1
        n_labels = size(labels)[2]

        for i in 1:n_labels
            label = view(labels, :, i)
            unique_label = filter(x -> !ismissing(x), unique(label))
            len_unique_label = length(unique_label)
            nclass = max(nclass, len_unique_label)
        end
        return nclass
    end

    function create_hypergraph(labels)
        H = nothing
        len_labels, n_labels = size(labels)

        for i in 1:n_labels
            label = view(labels, :, i)
            unique_label = filter(x -> !ismissing(x), unique(label))
            len_unique_label = length(unique_label)
            label2id = Dict(zip(unique_label, 1:len_unique_label))
            h = spzeros(len_labels, len_unique_label)
            for (i, elem) in enumerate(label)
                if ismissing(elem) == false
                    h[i, label2id[elem]] = 1.0
                end
            end
            H = H === nothing ? h : [H h]
        end
        return H
    end

    # Cluster-based Similarity Partitioning Algorithm
    function cspa(labels, nclass)
        H = create_hypergraph(labels)
        S = H * H'
        #label_ce = Metis.partition(Graph(S), nclass, alg=:RECURSIVE)
        label_ce = partition(SimpleWeightedGraph(S), nclass, alg=:RECURSIVE)
        label_ce = convert.(Int, label_ce)
        return label_ce
    end

    # HyperGraph Partitioning Algorithm
    function hgpa(labels, nclass)
        H = KaHyPar.HyperGraph(create_hypergraph(labels))
        label_ce = KaHyPar.partition(H, nclass, configuration=default_configuration)
        label_ce = convert.(Int, label_ce .+ 1)
        return label_ce
    end

    # Meta-CLustering Algorithm
    function mcla(labels, nclass)
        # Construct Meta-graph
        len_labels = size(labels)[1]
        H = create_hypergraph(labels)

        pair_dist_jac = pairwise(Jaccard(), H, H, dims=2)
        S = sparse(ones(size(pair_dist_jac)) - pair_dist_jac)
        S = convert.(Int, round.(S .* 1e3))

        # Cluster Hyperedges
        #membership = Metis.partition(Graph(S), nclass, alg=:RECURSIVE)
        membership = partition(SimpleWeightedGraph(S), nclass, alg=:RECURSIVE)

        # Collapse Meta-clusters
        meta_clusters = spzeros(len_labels, nclass)
        for (i, v) in enumerate(membership)
            for j in 1:len_labels
                meta_clusters[j, v] += H[j, i]
            end
        end

        # Compete for Objects
        label_ce = zeros(Int, len_labels)
        for i in 1:len_labels
            v = view(meta_clusters, i, :)
            candidate = []
            max_v = maximum(v)
            for (idx_v, elem_v) in enumerate(v)
                if elem_v == max_v
                    push!(candidate, idx_v)
                end
            end
            label_ce[i] = rand(candidate)
        end

        label_ce = convert.(Int, label_ce)
        return label_ce
    end

    # Hybrid Bipartite Graph Formulation
    function hbgf(labels, nclass)
        A = create_hypergraph(labels)
        n_rows, n_cols = size(A)

        W = [spzeros(n_cols, n_cols) A'; A spzeros(n_rows, n_rows)]
        membership = Metis.partition(Graph(W), nclass, alg=:RECURSIVE)
        label_ce = convert.(Int, membership[n_cols+1:end])

        return label_ce
    end

    function create_connectivity_matrix(labels)
        len_labels, n_labels = size(labels)
        M = zeros(len_labels, len_labels)
        m = zeros(len_labels, len_labels)

        for i in 1:n_labels
            label = view(labels, :, i)
            for j in 1:len_labels
                if ismissing(label[j])
                    continue
                end
                for k in 1:len_labels
                    if ismissing(label[k])
                        continue
                    end
                    if label[j] == label[k]
                        m[j, k] = 1
                    end
                end
            end
            M += m
            fill!(m, 0)
        end
        M ./= n_labels

        return M
    end

    function orthogonal_nmf_algorithm(W::Matrix{T}, nclass, maxiter) where T
        n = size(W)[1]
        Q = rand(n, nclass)
        S = diagm(rand(nclass))
        QS = Matrix{T}(undef, n, nclass)
        WQS = Matrix{T}(undef, n, nclass)
        QTQ = Matrix{T}(undef, nclass, nclass)

        for _ in 1:maxiter
            # Update Q
            mul!(QS, Q, S)
            mul!(WQS, W, QS)
            Q .*= sqrt.( WQS ./ (Q * Q' * WQS .+ 1e-8) )
            # Update S
            mul!(QTQ, Q', Q)
            S .*= sqrt.( (Q' * W * Q) ./ (QTQ * S * QTQ .+ 1e-8) )
        end

        return Q, S
    end

    # NMF-based consensus clustering
    function nmf(labels, nclass; maxiter=200)
        len_labels = size(labels)[1]

        M = create_connectivity_matrix(labels)
        Q, S = orthogonal_nmf_algorithm(M, nclass, maxiter)
        tmp = Q * sqrt.(S)

        label_ce = Array{Int}(undef, len_labels)
        for i in 1:len_labels
            label_ce[i] = argmax(view(tmp, i, :))
        end

        return label_ce
    end

    # Calculate the objective function value for cluster ensembles
    function calc_objective(labels, label_ce)
        objv = 0.0
        for i in 1:size(labels)[2]
            label = view(labels, :, i)
            idx = .!ismissing.(label)
            cc = convert(Array{Int}, label_ce[idx])
            label = convert(Array{Int}, label[idx])
            objv += mutualinfo(cc, label, normed=true)
        end
        return objv
    end

    """
    `cluster_ensembles` generate a single consensus clustering label by using 
    base labels obtained from multiple clustering algorithms.
    
    # Arguments
    - `labels`: The given set of labels obtained from multiple clustering algorithms. 
                Each column is a label.
    - `nclass`: Number of classes in consensus clustering label.
    - `random_state`: Used for reproducible results.
    - `alg`: Algorithms for cluster ensembles.
    """
    function cluster_ensembles(labels::Union{Array{Int}, Array{Union{Int, Missing}}}; 
                               nclass::Union{Int, Nothing}=nothing,
                               random_state::Union{Int, Nothing}=nothing, 
                               alg::Symbol=:hbgf)
        if nclass === nothing
            nclass = set_nclass(labels)
        end

        eltype(nclass) <: Number || nclass > 0 || throw(ArgumentError("nclass must be positive."))

        if random_state !== nothing
            eltype(random_state) <: Number || random_state >= 0 || throw(ArgumentError("random_state must be nonnegative."))
            Random.seed!(random_state)
        end

        if alg == :hbgf
            label_ce = hbgf(labels, nclass)
        elseif alg == :mcla
            label_ce = mcla(labels, nclass)
        elseif alg == :nmf
            label_ce = nmf(labels, nclass)
        elseif alg == :hgpa
            label_ce = hgpa(labels, nclass)
        elseif alg == :cspa
            if size(labels)[1] > 5000
                @warn "The length of labels is very large, so another solvers is recommended."
            end
            label_ce = cspa(labels, nclass)
        elseif alg == :all
            if size(labels)[1] > 5000
                algs = [hgpa, mcla, hbgf, nmf]
            else
                algs = [cspa, hgpa, mcla, hbgf, nmf]
            end
            best_objv = nothing
            for a in algs
                label = a(labels, nclass)
                objv = calc_objective(labels, label)
                if best_objv === nothing
                    best_objv = objv
                    label_ce = label
                end
                if best_objv < objv
                    best_objv = objv
                    label_ce = label
                end
            end
        else
            throw(ArgumentError("Invalid algorithm."))
        end

        return label_ce
    end
end
